"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tyr_1 = require("../tyr");
const collection_1 = require("./collection");
const query_1 = require("./query");
const Subscription = new collection_1.default({
    id: '_t3',
    name: 'tyrSubscription',
    client: false,
    internal: true,
    fields: {
        _id: { is: 'mongoid' },
        u: { link: 'user?', label: 'User' },
        c: { is: 'string', label: 'Collection' },
        q: { is: 'string', label: 'Query', note: 'Stringified MongoDB query.' },
        on: { is: 'date' },
        // TODO:  this is temporary, long-term would like to hook up tyranid to session table and use that to
        //        determine user -> instance bindings
        i: { is: 'string', label: 'Instance' },
    },
});
/*
interface LocalListener {
  [collectionId: string]: {
    changeHandlerDereg: () => void,
    removeHandlerDereg: () => void,
    queries: {
      [queryStr: string]: {
        queryObj: MongoDBQuery,
        instances: {
          [instanceId: string]: boolean
        },
        users: {
          [userId: string]: boolean
        }
      }
    }
  }
}
*/
let localListeners /*: LocalListener*/ = {};
async function fireEvent(colId, queryDef, refinedDocument, refinedQuery, type) {
    //con sole.log('fireEvent', colId, queryDef, refinedDocument, refinedQuery);
    if (tyr_1.default.logging.trace)
        tyr_1.default.trace({
            e: 'subscription',
            c: colId,
            m: 'matched',
            instances: queryDef.instances,
        });
    const promises = [];
    for (const instanceId in queryDef.instances) {
        const event = new tyr_1.default.Event({
            collection: Subscription,
            dataCollectionId: colId,
            query: refinedQuery,
            document: refinedDocument,
            type: 'subscriptionEvent',
            when: 'pre',
            instanceId: instanceId,
            subType: type,
        });
        if (instanceId === tyr_1.default.instanceId) {
            handleSubscriptionEvent(event);
        }
        else {
            promises.push(tyr_1.default.Event.fire(event));
        }
    }
    return Promise.all(promises);
}
async function parseSubscriptions(subscription, userId) {
    /*con sole.log(
      'parseSubscriptions(), Tyr.instanceId=',
      Tyr.instanceId,
      'userId=',
      userId
    );*/
    let subs;
    if (subscription) {
        subs = [subscription];
    }
    else if (userId) {
        // clear out existing data for this userId
        subs = await Subscription.findAll({ query: { u: userId } });
        for (const colId in localListeners) {
            const queries = localListeners[colId].queries;
            for (const queryStr in queries) {
                delete queries[queryStr].users[userId];
                //if (queryDef.users is empty) then clear out queryDef.instances ?
            }
        }
    }
    else {
        subs = await Subscription.findAll({});
        // if we're reparsing all subs, clear out existing data
        for (const colId in localListeners) {
            const listener = localListeners[colId];
            if (listener) {
                listener.changeHandlerDereg && listener.changeHandlerDereg();
                listener.removeHandlerDereg && listener.removeHandlerDereg();
            }
        }
        localListeners = {};
    }
    //con sole.log(Tyr.instanceId + ' *** parseSubscriptions, ' + subs.length + ' subs');
    for (const sub of subs) {
        const colId = sub.c, col = tyr_1.default.byId[colId];
        if (!col) {
            console.warn(`Existing subscription to collection id ${colId} which no longer exists -- skipping`);
            continue;
        }
        let listener = localListeners[colId];
        if (!localListeners[colId]) {
            const changeHandlerDereg = col.on({
                type: 'change',
                when: 'post',
                handler: async (event) => {
                    //con sole.log(Tyr.instanceId + ' *** ' + col.def.name + ' change:'); //, event);
                    const { document, query, _documents } = event;
                    const promises = [];
                    for (const queryStr in listener.queries) {
                        const queryDef = listener.queries[queryStr];
                        if (document) {
                            if (query_1.default.matches(queryDef.queryObj, document)) {
                                promises.push(fireEvent(colId, queryDef, document, undefined, event.type));
                            }
                        }
                        else if (_documents) {
                            promises.push(..._documents
                                .filter(doc => query_1.default.matches(queryDef.queryObj, doc))
                                .map(doc => fireEvent(colId, queryDef, doc, undefined, event.type)));
                        }
                        else {
                            /*if (query)*/ const refinedQuery = query_1.default.intersection(queryDef.queryObj, query);
                            if (refinedQuery) {
                                promises.push(fireEvent(colId, queryDef, null, refinedQuery, event.type));
                            }
                        }
                    }
                    await Promise.all(promises);
                },
            });
            const removeHandlerDereg = col.on({
                type: 'remove',
                when: 'post',
                handler: async (event) => {
                    //con sole.log(Tyr.instanceId + ' *** ' + col.def.name + ' remove:'); //, event);
                    const { document, query, _documents } = event;
                    const promises = [];
                    for (const queryStr in listener.queries) {
                        const queryDef = listener.queries[queryStr];
                        if (document) {
                            if (query_1.default.matches(queryDef.queryObj, document)) {
                                promises.push(fireEvent(colId, queryDef, document, undefined, 'remove'));
                            }
                        }
                        else if (_documents) {
                            promises.push(..._documents
                                .filter(doc => query_1.default.matches(queryDef.queryObj, doc))
                                .map(doc => fireEvent(colId, queryDef, doc, undefined, 'remove')));
                        }
                        else {
                            /*if (query)*/ const refinedQuery = query_1.default.intersection(queryDef.queryObj, query);
                            if (refinedQuery) {
                                promises.push(fireEvent(colId, queryDef, null, refinedQuery, 'remove'));
                            }
                        }
                    }
                    await Promise.all(promises);
                },
            });
            listener = localListeners[colId] = {
                changeHandlerDereg,
                removeHandlerDereg,
                queries: {},
            };
        }
        const queryStr = sub.q;
        let queryDef = listener.queries[queryStr];
        if (!queryDef) {
            queryDef = listener.queries[queryStr] = {
                queryObj: col.fromClientQuery(JSON.parse(queryStr)),
                instances: {},
                users: {},
            };
        }
        queryDef.instances[sub.i] = true;
        queryDef.users[sub.u] = true;
    }
}
Subscription.on({
    type: 'subscribe',
    when: 'post',
    async handler(event) {
        await parseSubscriptions(event.subscription);
    },
});
Subscription.on({
    type: 'unsubscribe',
    when: 'post',
    async handler(event) {
        await parseSubscriptions(undefined, event.user);
    },
});
//let bootNeeded = 'Subscription needs to be booted';
Subscription.boot = async function (stage, pass) {
    if (stage === 'link' && tyr_1.default.db) {
        //if (bootNeeded) {
        await parseSubscriptions();
        //bootNeeded = undefined;
        //}
        return undefined; //bootNeeded;
    }
};
collection_1.default.prototype.subscribe = async function (query, user, cancel) {
    if (tyr_1.default.logging.trace) {
        tyr_1.default.trace({
            c: this.id,
            e: 'subscription',
            m: 'subscribe:' + (user ? user._id : 'no user'),
            q: query,
            cancel,
        });
    }
    if (!user) {
        console.warn(`${this.name}.subscribe() called without a user -- ignoring`);
        return;
    }
    const queryStr = JSON.stringify(query);
    if (!query) {
        if (cancel) {
            await Subscription.remove({ query: { u: user._id, c: this.id } });
            await tyr_1.default.Event.fire({
                collection: Subscription,
                type: 'unsubscribe',
                when: 'post',
                broadcast: true,
                user: user._id,
            });
            return;
        }
        else {
            throw new Error('missing query');
        }
    }
    const subscription = await Subscription.findOne({
        query: {
            u: user._id,
            c: this.id,
            q: queryStr,
        },
    });
    if (cancel) {
        if (subscription) {
            await subscription.$remove();
            await tyr_1.default.Event.fire({
                collection: Subscription,
                type: 'unsubscribe',
                when: 'post',
                broadcast: true,
                user: user._id,
            });
        }
        return;
    }
    if (!subscription || subscription.i !== tyr_1.default.instanceId) {
        let s = subscription;
        if (s) {
            delete s._id;
            s.on = new Date();
            s.i = tyr_1.default.instanceId;
        }
        else {
            s = new Subscription({
                u: user._id,
                c: this.id,
                q: queryStr,
                on: new Date(),
                i: tyr_1.default.instanceId,
            });
        }
        await tyr_1.default.Event.fire({
            collection: this,
            type: 'subscribe',
            when: 'pre',
            query,
            opts: {
                query,
                auth: user,
            },
            subscription: s,
        });
        await s.$save();
        await tyr_1.default.Event.fire({
            collection: Subscription,
            type: 'subscribe',
            when: 'post',
            broadcast: true,
            subscription: s,
        });
    }
};
Subscription.unsubscribe = async function (userId) {
    const rslts = await Subscription.remove({
        query: {
            u: userId,
        },
    });
    if (rslts.result.n) {
        await tyr_1.default.Event.fire({
            collection: Subscription,
            type: 'unsubscribe',
            when: 'post',
            broadcast: true,
            user: userId,
        });
    }
};
async function handleSubscriptionEvent(event) {
    //con sole.log(Tyr.instanceId + ' *** handleSubscriptionEvent:'); //, event);
    const col = event.dataCollection, listener = localListeners[col.id], mQuery = event.query, mDoc = event.document;
    if (listener) {
        const userIds = {};
        for (const queryStr in listener.queries) {
            const queryDef = listener.queries[queryStr];
            if (mQuery) {
                if (query_1.default.intersection(queryDef.queryObj, mQuery)) {
                    for (const userId in queryDef.users) {
                        userIds[userId] = true;
                    }
                }
            }
            else {
                // if mDoc
                if (query_1.default.matches(queryDef.queryObj, mDoc)) {
                    for (const userId in queryDef.users) {
                        userIds[userId] = true;
                    }
                }
            }
        }
        const documents = await event.documents;
        const sockets = tyr_1.default.io.sockets.sockets;
        for (const socketId in sockets) {
            const socket = sockets[socketId];
            if (userIds[socket.tyrUserId]) {
                socket.emit('subscriptionEvent', {
                    colId: event.dataCollectionId,
                    type: event.subType,
                    docs: documents.map(doc => doc.$toClient()),
                });
            }
        }
    }
}
Subscription.on({
    type: 'subscriptionEvent',
    handler: handleSubscriptionEvent,
});
exports.default = Subscription;
//# sourceMappingURL=subscription.js.map